'''Show basic usage of GS solution.'''

import numpy as np
import matplotlib.pyplot as plt
from phantominator import shepp_logan
from ssfp import bssfp, gs_recon

if __name__ == '__main__':

    # Shepp-Logan
    N = 128
    M0 = shepp_logan(N)
    T1, T2 = M0*2, M0/2

    # Simulate bSSFP acquisition with linear off-resonance
    TR, alpha = 3e-3, np.deg2rad(30)
    pcs = np.linspace(0, 2*np.pi, 4, endpoint=False)
    df, _ = np.meshgrid(
        np.linspace(-1/TR, 1/TR, N),
        np.linspace(-1/TR, 1/TR, N))
    sig = bssfp(T1, T2, TR, alpha, field_map=df, phase_cyc=pcs, M0=M0)

    # Show the phase-cycled images
    nx, ny = 2, 2
    plt.figure()
    for ii in range(nx*ny):
        plt.subplot(nx, ny, ii+1)
        plt.imshow(np.abs(sig[ii, ...]))
        plt.title('%d deg PC' % (ii*90))
    plt.show(block=False)

    # Dhow the recon
    recon = gs_recon(sig, pc_axis=0)
    plt.figure()
    plt.imshow(np.abs(recon))
    plt.title('GS Solution')
    plt.show()



'''bSSFP simulation.'''

import numpy as np

def bssfp(
        T1, T2, TR, alpha, field_map=0, phase_cyc=0, M0=1, delta_cs=0,
        phi_rf=0, phi_edd=0, phi_drift=0):
    r'''bSSFP transverse signal at time TE after excitation.

    Parameters
    ----------
    T1 : float or array_like
        longitudinal exponential decay time constant (in seconds).
    T2 : float or array_like
        transverse exponential decay time constant (in seconds).
    TR : float
        repetition time (in seconds).
    alpha : float or array_like
        flip angle (in rad).
    field_map : float or array_like, optional
        B0 field map (in Hz).
    phase_cyc : float or array_like, optional
        Linear phase-cycle increment (in rad).
    M0 : float or array_like, optional
        proton density.
    delta_cs : float, optional
        chemical shift of species w.r.t. the water peak (in Hz).
    phi_rf : float, optional
        RF phase offset, related to the combin. of Tx/Rx phases (in
        rad).
    phi_edd : float, optional
        phase errors due to eddy current effects (in rad).
    phi_drift : float, optional
        phase errors due to B0 drift (in rad).

    Returns
    -------
    Mxy : numpy.array
        Transverse complex magnetization.

    Notes
    -----
    `T1`, `T2`, `alpha`, `field_map`, and `M0` can all be either a
    scalar or an MxN array.  `phase_cyc` can be a scalar or length L
    vector.

    Implementation of equations [1--2] in [1]_.  These equations are
    based on the Ernst-Anderson derivation [4]_ where off-resonance
    is assumed to be subtracted as opposed to added (as in the
    Freeman-Hill derivation [5]_).  Hoff actually gets Mx and My
    flipped in the paper, so we fix that here.  We also assume that
    the field map will be provided given the Freeman-Hill convention.

    We will additionally assume that linear phase increments
    (phase_cyc) will be given in the form:

    .. math::

        \theta = 2 \pi (\delta_{cs} + \Delta f_0)\text{TR} + \Delta
        \theta.

    Notice that this is opposite of the convention used in PLANET,
    where phase_cyc is subtracted (see equation [12] in [2]_).

    Also see equations [2.7] and [2.10a--b] from [4]_ and equations
    [3] and [6--12] from [5]_.

    References
    ----------
    .. [1] Xiang, Qing‐San, and Michael N. Hoff. "Banding artifact
           removal for bSSFP imaging with an elliptical signal
           model." Magnetic resonance in medicine 71.3 (2014):
           927-933.

    .. [4] Ernst, Richard R., and Weston A. Anderson. "Application of
           Fourier transform spectroscopy to magnetic resonance."
           Review of Scientific Instruments 37.1 (1966): 93-102.

    .. [5] Freeman R, Hill H. Phase and intensity anomalies in
           fourier transform NMR. J Magn Reson 1971;4:366–383.
    '''

    # We are assuming Freeman-Hill convention for off-resonance map,
    # so we need to negate to make use with this Ernst-Anderson-
    # based implementation from Hoff
    field_map = -1*field_map

    # We also assume that linear phase cycles will be added, but the
    # formulation used by Hoff, PLANET assumes subtracted, so let's
    # flip the signs
    phase_cyc = -1*phase_cyc

    # Make sure we're working with arrays
    T1 = np.atleast_2d(T1)
    T2 = np.atleast_2d(T2)
    alpha = np.atleast_2d(alpha)
    field_map = np.atleast_2d(field_map)
    phase_cyc = np.atleast_2d(phase_cyc)

    # If we have more than one phase-cycle, then add that dimension
    if phase_cyc.size > 1:
        reps = (phase_cyc.size, 1, 1)
        phase_cyc = np.tile(
            phase_cyc, T1.shape[:] + (1,)).transpose((2, 0, 1))
        T1 = np.tile(T1, reps)
        T2 = np.tile(T2, reps)
        alpha = np.tile(alpha, reps)
        field_map = np.tile(field_map, reps)

    # All this nonsense so we don't divide by 0
    E1 = np.zeros(T1.shape)
    E1[T1 > 0] = np.exp(-TR/T1[T1 > 0])
    E2 = np.zeros(T2.shape)
    E2[T2 > 0] = np.exp(-TR/T2[T2 > 0])

    # Precompute theta and some cos, sin
    theta = _get_theta(TR, field_map, phase_cyc, delta_cs)
    ca = np.cos(alpha)
    sa = np.sin(alpha)
    ct = np.cos(theta)
    st = np.sin(theta)

    # Get to business
    den = (1 - E1*ca)*(1 - E2*ct) - (E2*(E1 - ca))*(E2 - ct)
    Mx = -M0*((1 - E1)*E2*sa*st)/den
    My = M0*((1 - E1)*sa)*(1 - E2*ct)/den
    Mxy = Mx + 1j*My

    # Add additional phase factor for readout at TE = TR/2.
    # Notice that phi_i are negated
    Mxy *= _get_bssfp_phase(
        T2, TR, field_map, delta_cs, -phi_rf, -phi_edd, -phi_drift)

    return Mxy.squeeze()

def _get_bssfp_phase(
        T2, TR, field_map, delta_cs=0, phi_rf=0, phi_edd=0,
        phi_drift=0):
    '''Additional bSSFP phase factors.

    Parameters
    ----------
    T2 : array_like
        Longitudinal relaxation constant (in sec).
    TR : float
        Repetition time (in sec).
    field_map : array_like
        off-resonance map (Hz).
    delta_cs : float, optional
        chemical shift of species w.r.t. the water peak (Hz).
    phi_rf : float, optional
        RF phase offset, related to the combin. of Tx/Rx phases (rad).
    phi_edd : float, optional
        phase errors due to eddy current effects (rad).
    phi_drift : float, optional
        phase errors due to B0 drift (rad).

    Returns
    -------
    phase : array_like
        Additional phase term to simulate readout at time TE = TR/2.
        Assumes balanced (TE = TR/2).

    Notes
    -----
    This is exp(-i phi) from end of p. 930 in [1]_.

    We use a positive exponent, exp(i phi), as in Hoff and Taylor
    MATLAB implementations.  This phase factor is also positive in
    equaiton [5] of [3]_.

    In Hoff's paper the equation is not explicitly given for phi, so
    we implement equation [5] that gives more detailed terms, found
    in [2]_.

    References
    ----------
    .. [2] Shcherbakova, Yulia, et al. "PLANET: An ellipse fitting
           approach for simultaneous T1 and T2 mapping using
           phase‐cycled balanced steady‐state free precession."
           Magnetic resonance in medicine 79.2 (2018): 711-722.

    .. [3] Scheffler, Klaus, and Jürgen Hennig. "Is TrueFISP a
           gradient‐echo or a spin‐echo sequence?." Magnetic
           Resonance in Medicine: An Official Journal of the
           International Society for Magnetic Resonance in Medicine
           49.2 (2003): 395-397.
    '''

    TE = TR/2 # assume bSSFP
    phi = 2*np.pi*(
        delta_cs + field_map)*TE + phi_rf + phi_edd + phi_drift

    T2 = np.array(T2)
    idx = np.where(T2 > 0)
    val = np.zeros(T2.shape)
    val[idx] = -TE/T2[idx]
    return np.exp(1j*phi)*np.exp(val)

def _get_theta(TR, field_map, phase_cyc=0, delta_cs=0):
    '''Get theta, spin phase per repetition time, given off-resonance.

    Parameters
    ----------
    TR : float
        repetition time (in sec).
    field_map : array_like
        Off-resonance map (in Hz).
    phase_cyc : array_like, optional
        Phase-cycling (in rad).
    delta_cs : float, optional, optional
        Chemical shift of species w.r.t. the water peak (Hz).

    Returns
    -------
    theta : array_like
        Spin phase per repetition time, given off-resonance.

    Notes
    -----
    Equation for theta=2*pi*df*TR is in Appendix A of [6]_.  The
    additional chemical shift term can be found, e.g., in [2]_.

    References
    ----------
    .. [6] Hargreaves, Brian A., et al. "Characterization and
           reduction of the transient response in steady‐state MR
           imaging." Magnetic Resonance in Medicine: An Official
           Journal of the International Society for Magnetic
           Resonance in Medicine 46.1 (2001): 149-158.
    '''

    return 2*np.pi*(delta_cs + field_map)*TR + phase_cyc

if __name__ == '__main__':
    pass



'''Basic usage of bSSFP.'''

import numpy as np
import matplotlib.pyplot as plt
from ssfp import bssfp

if __name__ == '__main__':

    M0, T1, T2 = 1, 1, .5
    TR, alpha = 3e-3, np.deg2rad(30)
    field_map = 1/(2*TR)
    pcs = np.linspace(0, 2*np.pi, 100, endpoint=False)
    sig = bssfp(
        T1, T2, TR, alpha, field_map, phase_cyc=pcs, M0=M0,
        delta_cs=0, phi_rf=0, phi_edd=0, phi_drift=0)

    fig, ax1 = plt.subplots()
    dpcs = np.rad2deg(pcs)
    ax1.plot(dpcs, np.abs(sig))
    ax1.set_ylabel('Magnitude (a.u.)')
    ax1.set_xlabel('Phase Cycle (in deg)')

    ax2 = ax1.twinx()
    ax2.plot(dpcs, np.angle(sig), '--')
    ax2.set_ylabel('Phase (rad)')

    plt.title('bSSFP Spectrum')
    plt.show()



'''Setup.py'''

from distutils.core import setup
from setuptools import find_packages

setup(
    name='ssfp',
    version='0.1.0',
    author='Nicholas McKibben',
    author_email='nicholas.bgp@gmail.com',
    packages=find_packages(),
    scripts=[],
    url='https://github.com/mckib2/ssfp',
    license='GPLv3',
    description='SSFP simulation',
    long_description=open('README.rst').read(),
    install_requires=[
        "numpy>=1.17.2",
        "matplotlib>=3.1.1",
        "scikit-image>=0.15.0",
        "phantominator>=0.4.3",
    ],
    python_requires='>=3.6',
)



'''Geometric solution to the elliptical signal model.'''

import numpy as np
from skimage.util.shape import view_as_windows

def gs_recon(
        Is, pc_axis=0, isophase=np.pi, second_pass=True,
        patch_size=None):
    '''Full 2D Geometric Solution following Xiang, Hoff's 2014 paper.

    Parameters
    ----------
    Is : array_like
        4 phase-cycled images: [I0, I1, I2, I3].  (I0, I2) and
        (I1, I3) are phase-cycle pairs.
    pc_axis : int, optional
        Phase-cycle dimension, default is the first dimension.
    isophase : float
        Only neighbours with isophase max phase difference contribute.
    second_pass : bool, optional
        Compute the second pass solution, increasing SNR by sqrt(2).
    patch_size : tuple, optional
        Size of patches in pixels (x, y).

    Returns
    -------
    I : array_like
        Second pass GS solution to elliptical signal model.
    Id : array_like, optional
        If second_pass=False, GS solution to the elliptical signal
        model (without linear weighted solution).

    Notes
    -----
    `Is` is an array of 4 2D images: I0, I1, I2, and I3.  I0 and I2
    make up the first phase-cycle pair, that is they are 180 degrees
    phase-cycled relative to each other.  I1 and I3 are also
    phase-cycle pairs and must be different phase-cycles than either
    I0 or I2.  Relative phase-cycles are assumed as follows:

        - I0: 0 deg
        - I1: 90 deg
        - I2: 180 deg
        - I3: 270 deg

    Implements algorithm shown in Fig 2 of [1]_.

    References
    ----------
    .. [1] Xiang, Qing‐San, and Michael N. Hoff. "Banding artifact
           removal for bSSFP imaging with an elliptical signal
           model." Magnetic resonance in medicine 71.3 (2014):
           927-933.
    '''

    # Put the pc_axis first
    Is = np.moveaxis(Is, pc_axis, 0)

    # Get direct geometric solution for demoduled M for all pixels
    Id = np.atleast_1d(_get_complex_cross_point(Is))

    # Get maximum pixel magnitudes for all input images
    I_max_mag = np.atleast_1d(np.max(np.abs(Is), axis=0))

    # Compute complex sum
    CS = np.atleast_1d(np.mean(Is, axis=0))

    # For each pixel, if the magnitude if greater than the maximum
    # magnitude of all four input images, then replace the pixel with
    # the CS solution.  This step regularizes the direct solution and
    # effectively removes all singularities
    mask = np.abs(Id) > I_max_mag
    Id[mask] = CS[mask]

    # Bail early if we don't want the second pass solution
    if not second_pass:
        return Id

    # Find weighted sums of image pairs (I1,I3) and (I2,I4)
    Iw13 = _compute_Iw(
        Is[0, ...], Is[2, ...], Id, patch_size=patch_size,
        isophase=isophase)
    Iw24 = _compute_Iw(
        Is[1, ...], Is[3, ...], Id, patch_size=patch_size,
        isophase=isophase)

    # Final result is found by averaging the two linear solutions for
    # reduced noise
    return (Iw13 + Iw24)/2

def _get_complex_cross_point(Is):
    '''Find intersection of two lines connecting diagonal pairs.

    Parameters
    ----------
    Is : array_like
        4 phase-cycled images: [I0, I1, I2, I3].

    Returns
    -------
    M : array_like
        Complex cross point.

    Notes
    -----
    We assume that Is has the phase-cycle dimenension along the first
    axis.

    (xi, yi) are the real and imaginary parts of complex valued
    pixels in four bSSFP images denoted Ii and acquired with phase
    cycling dtheta = (i-1)*pi/2 with 0 < i < 4.

    This is Equation [13] from [1]_.
    '''

    x1, y1 = Is[0, ...].real, Is[0, ...].imag
    x2, y2 = Is[1, ...].real, Is[1, ...].imag
    x3, y3 = Is[2, ...].real, Is[2, ...].imag
    x4, y4 = Is[3, ...].real, Is[3, ...].imag

    den = (x1 - x3)*(y2 - y4) + (x2 - x4)*(y3 - y1)
    if (den == 0).any():
        # Make sure we're not dividing by zero
        den += np.finfo(float).eps

    M = ((x1*y3 - x3*y1)*(Is[1, ...] - Is[3, ...]) - (
        x2*y4 - x4*y2)*(Is[0, ...] - Is[2, ...]))/den
    return M

def _compute_Iw(
        I0, I1, Id, patch_size=None, mode='constant', isophase=np.pi,
        ret_weight=False):
    '''Computes weighted sum of image pair (I0, I1).

    Parameters
    ----------
    I0 : array_like
        1st of pair of diagonal images (relative phase cycle of 0).
    I1 : array_like
        2nd of pair of diagonal images (relative phase cycle of 180
        deg).
    Id : array_like
        result of regularized direct solution.
    patch_size : tuple, optional
        size of patches in pixels (x, y).  Defaults to (5, 5).
    mode : {'contant', 'edge'}, optional
        mode of numpy.pad. Probably choose 'constant' or 'edge'.
    isophase : float, optional
        Only neighbours with isophase max phase difference contribute.
    ret_weight : bool, optional
        Return weight, w.

    Returns
    -------
    Iw : array_like
        The weighted sum of image pair (I0,I1), equation [14]
    w : array_like, optional
        If ret_weight=True, w is returned.

    Notes
    -----
    Image pair (I0,I1) are phase cycled bSSFP images that are
    different by 180 degrees.  Id is the image given by the direct
    method (Equation [13]) after regularization by the complex sum.
    This function solves for the weights by regional differential
    energy minimization.  The 'regional' part means that the image is
    split into patches of size patch_size with edge boundary
    conditions (pads with the edge values given by mode option).  The
    weighted sum of the image pair is returned.

    The isophase does not appear in the paper, but appears in Hoff's
    MATLAB code.  It appears that we only want to consider pixels in
    the patch that have similar tissue properties - in other words,
    have similar phase.  The default isophase is pi as in Hoff's
    implementation.

    This function implements Equations [14,18], or steps 4--5 from
    Fig. 2 in [1]_.
    '''

    # Make sure we have a patch size
    if patch_size is None:
        patch_size = (5, 5)

    # Expressions for the numerator and denominator
    numerator = np.conj(I1 - Id)*(I1 - I0) + np.conj(I1 - I0)*(
        I1 - Id)
    den = np.conj(I0 - I1)*(I0 - I1)

    # We'll have trouble with a 1d input if we don't do this
    numerator = np.atleast_2d(numerator)
    den = np.atleast_2d(den)

    # Pad the image so we can generate patches where we need them
    edge_pad = [int(p/2) for p in patch_size]
    numerator = np.pad(numerator, pad_width=edge_pad, mode=mode)
    den = np.pad(den, pad_width=edge_pad, mode=mode)

    # Separate out into patches of size patch_size
    numerator_patches = view_as_windows(numerator, patch_size)
    den_patches = view_as_windows(den, patch_size)

    # Make sure the phase difference is below a certan bound to
    # include point in weights
    mask = _mask_isophase(numerator_patches, patch_size, isophase)
    numerator_patches *= mask
    den_patches *= mask

    numerator_weights = np.sum(numerator_patches, axis=(-2, -1))
    den_weights = np.sum(den_patches, axis=(-2, -1))

    # Equation [18]
    weights = numerator_weights/(2*den_weights + np.finfo(float).eps)

    # Find Iw, the weighted sum of image pair (I0,I1), equation [14]
    Iw = I0*weights + I1*(1 - weights)
    if ret_weight:
        return(Iw, weights)
    return Iw

def _mask_isophase(numerator_patches, patch_size, isophase):
    '''Generate mask that chooses patch pixels that satisfy isophase.

    Parameters
    ----------
    numerator_patches : array_like
        Numerator patches from second pass solution.
    patch_size : tuple
        size of patches in pixels (x,y).
    isophase : float
        Only neighbours with isophase max phase difference contribute.

    Returns
    -------
        mask : array_like
            same size as numerator_patches, to be applied to
            numerator_patches and den_patches before summation.
    '''

    # # Loop through each patch and zero out all the values not
    # mask = np.ones(num_patches.shape).astype(bool)
    # center_x,center_y = int(patch_size[0]/2),int(patch_size[1]/2)
    # for ii in range(mask.shape[0]):
    #     for jj in range(mask.shape[1]):
    #         mask[ii,jj,...] = np.abs(np.angle(
    #            num_patches[ii,jj,...])*np.conj(
    #                num_patches[ii,jj,center_x,center_y])) < isophase
    # # print(np.sum(mask == False))

    # Now try it without loops - it'll be faster...
    center_x, center_y = [int(p/2) for p in patch_size]
    ref_pixels = np.repeat(np.repeat(
        numerator_patches[:, :, center_x, center_y, None],
        patch_size[0], axis=-1)[..., None], patch_size[1], axis=-1)
    mask_mat = np.abs(np.angle(
        numerator_patches)*np.conj(ref_pixels)) < isophase
    # assert np.allclose(mask_mat,mask)

    return mask_mat


if __name__ == '__main__':
    pass
